@@ -318,7 +318,7 @@ opf_xml_text_handler (GMarkupParseContext   *context,
 				}
 			}
 
-			if (i == len) {
+			if (!fname && i == len) {
 				fname = g_strdup (data->savedstring);
 				g_debug ("Found only one name");
 			} else {
@@ -666,7 +666,8 @@ extract_opf (const gchar          *uri,
 }
 
 G_MODULE_EXPORT gboolean
-tracker_extract_get_metadata (TrackerExtractInfo *info)
+tracker_extract_get_metadata (TrackerExtractInfo  *info,
+                              GError             **error)
 {
 	TrackerResource *ebook;
 	gchar *opf_path, *uri;
