@@ -209,7 +209,8 @@ extract_ps_gz (const gchar *uri)
 #endif /* USING_UNZIPPSFILES */
 
 G_MODULE_EXPORT gboolean
-tracker_extract_get_metadata (TrackerExtractInfo *info)
+tracker_extract_get_metadata (TrackerExtractInfo  *info,
+                              GError             **error)
 {
 	TrackerResource *metadata;
 	GFile *file;
